package ru.dbotthepony.kstarbound.network

import io.netty.buffer.ByteBuf
import io.netty.buffer.ByteBufInputStream
import io.netty.buffer.ByteBufOutputStream
import io.netty.channel.ChannelDuplexHandler
import io.netty.channel.ChannelHandlerContext
import io.netty.channel.ChannelPromise
import it.unimi.dsi.fastutil.bytes.ByteArrayList
import it.unimi.dsi.fastutil.ints.Int2ObjectArrayMap
import it.unimi.dsi.fastutil.io.FastByteArrayInputStream
import it.unimi.dsi.fastutil.io.FastByteArrayOutputStream
import it.unimi.dsi.fastutil.objects.Reference2ObjectOpenHashMap
import org.apache.logging.log4j.LogManager
import ru.dbotthepony.kommons.io.readSignedVarInt
import ru.dbotthepony.kommons.io.readVarInt
import ru.dbotthepony.kommons.io.writeSignedVarInt
import ru.dbotthepony.kommons.io.writeVarInt
import ru.dbotthepony.kstarbound.client.network.packets.ForgetChunkPacket
import ru.dbotthepony.kstarbound.client.network.packets.ChunkCellsPacket
import ru.dbotthepony.kstarbound.client.network.packets.ForgetEntityPacket
import ru.dbotthepony.kstarbound.client.network.packets.JoinWorldPacket
import ru.dbotthepony.kstarbound.client.network.packets.SpawnWorldObjectPacket
import ru.dbotthepony.kstarbound.network.packets.serverbound.ClientConnectPacket
import ru.dbotthepony.kstarbound.network.packets.clientbound.ConnectSuccessPacket
import ru.dbotthepony.kstarbound.network.packets.clientbound.HandshakeChallengePacket
import ru.dbotthepony.kstarbound.network.packets.serverbound.HandshakeResponsePacket
import ru.dbotthepony.kstarbound.network.packets.ProtocolRequestPacket
import ru.dbotthepony.kstarbound.network.packets.ProtocolResponsePacket
import ru.dbotthepony.kstarbound.network.packets.clientbound.ServerDisconnectPacket
import ru.dbotthepony.kstarbound.network.packets.clientbound.WorldStartPacket
import ru.dbotthepony.kstarbound.server.network.packets.TrackedPositionPacket
import ru.dbotthepony.kstarbound.server.network.packets.TrackedSizePacket
import java.io.BufferedInputStream
import java.io.DataInputStream
import java.io.DataOutputStream
import java.io.InputStream
import java.util.zip.InflaterInputStream
import kotlin.math.absoluteValue
import kotlin.reflect.KClass

class PacketRegistry(val isLegacy: Boolean) {
	private val packets = ArrayList<Type<*>?>()
	private val missingNames = Int2ObjectArrayMap<String>()
	private val clazz2Type = Reference2ObjectOpenHashMap<KClass<*>, Type<*>>()

	private data class Type<T : IPacket>(val id: Int, val type: KClass<T>, val factory: IPacketReader<T>, val direction: PacketDirection)

	val size: Int
		get() = packets.size

	private fun <T : IPacket> add(type: KClass<T>, reader: IPacketReader<T>, direction: PacketDirection = PacketDirection.get(type)): PacketRegistry {
		if (packets.size >= 255)
			throw IndexOutOfBoundsException("Unable to add any more packet types! 255 is the max")

		if (type in clazz2Type)
			throw IllegalArgumentException("Already has packet handler for type $reader (${type})")

		val ptype = Type(packets.size, type, reader, direction)
		packets.add(ptype)
		clazz2Type[type] = ptype
		return this
	}

	private inline fun <reified T : IPacket> add(reader: IPacketReader<T>, direction: PacketDirection = PacketDirection.get(T::class)): PacketRegistry {
		return add(T::class, reader, direction)
	}

	private fun skip(amount: Int = 1) {
		for (i in 0 until amount) {
			packets.add(null)
		}
	}

	private fun skip(name: String) {
		missingNames[packets.size] = name
		packets.add(null)
	}

	inner class Serializer(val side: ConnectionSide) : ChannelDuplexHandler() {
		private val backlog = ByteArrayList()
		private var discardBytes = 0
		private var readableBytes = 0
		private var isCompressed = false
		private var readingType: Type<*>? = null

		override fun channelRead(ctx: ChannelHandlerContext, msg: Any) {
			if (msg is ByteBuf) {
				try {
					while (msg.readableBytes() > 0) {
						if (discardBytes > 0) {
							val toSkip = discardBytes.coerceAtMost(msg.readableBytes())
							discardBytes -= toSkip
							msg.skipBytes(toSkip)
							continue
						}

						if (readingType != null) {
							while (readableBytes > 0 && msg.readableBytes() > 0) {
								backlog.add(msg.readByte())
								readableBytes--
							}

							if (readableBytes == 0) {
								val stream: InputStream

								if (isCompressed) {
									stream = BufferedInputStream(InflaterInputStream(FastByteArrayInputStream(backlog.elements(), 0, backlog.size)))
								} else {
									stream = FastByteArrayInputStream(backlog.elements(), 0, backlog.size)
								}

								try {
									ctx.fireChannelRead(readingType!!.factory.read(DataInputStream(stream), isLegacy))
								} catch (err: Throwable) {
									LOGGER.error("Error while reading incoming packet from network (type ${readingType!!.id}; ${readingType!!.type})", err)
								}

								backlog.clear()
								readingType = null
								isCompressed = false
							}

							continue
						}

						val stream = DataInputStream(ByteBufInputStream(msg))
						val packetType = stream.readUnsignedByte()
						val dataLength = stream.readSignedVarInt()

						val type = packets.getOrNull(packetType)

						if (type == null) {
							val name = missingNames[packetType]

							if (name != null)
								LOGGER.error("Unknown packet type $packetType ($name)! Discarding ${dataLength.absoluteValue} bytes")
							else
								LOGGER.error("Unknown packet type $packetType! Discarding ${dataLength.absoluteValue} bytes")

							discardBytes = dataLength.absoluteValue
						} else if (!type.direction.acceptedOn(side)) {
							LOGGER.error("Packet type $packetType (${type.type}) can not be accepted on side $side! Discarding ${dataLength.absoluteValue} bytes")
							discardBytes = dataLength.absoluteValue
						} else {
							LOGGER.debug("Packet type {} ({}) received on {} (size {} bytes)", packetType, type.type, side, dataLength.absoluteValue)
							readingType = type
							readableBytes = dataLength.absoluteValue
							isCompressed = dataLength < 0
						}
					}
				} finally {
					msg.release()
				}
			} else {
				super.channelRead(ctx, msg)
			}
		}

		override fun write(ctx: ChannelHandlerContext, msg: Any, promise: ChannelPromise) {
			val type = clazz2Type[msg::class]

			if (type == null) {
				LOGGER.error("Unknown outgoing message type ${msg::class}, it will not reach the other side.")
			} else if (!type.direction.acceptedOn(side.opposite)) {
				LOGGER.error("Packet ${type.type} can not be accepted on side ${side.opposite}, refusing to send it!")
			} else {
				val stream = FastByteArrayOutputStream()
				(msg as IPacket).write(DataOutputStream(stream), isLegacy)

				if (isLegacy)
					check(stream.length > 0) { "Packet $msg didn't write any data to network, this is not allowed by legacy protocol" }

				val buff = ctx.alloc().buffer(stream.length + 5)
				val stream2 = ByteBufOutputStream(buff)
				stream2.writeByte(type.id)
				stream2.writeSignedVarInt(stream.length)
				stream2.write(stream.array, 0, stream.length)
				LOGGER.debug("Packet type {} ({}) sent from {} (size {} bytes)", type.id, type.type, side, stream.length)
				ctx.write(buff, promise)
			}
		}
	}

	inner class Validator(val side: ConnectionSide) : ChannelDuplexHandler() {
		override fun channelRead(ctx: ChannelHandlerContext, msg: Any) {
			val type = clazz2Type[msg::class]

			if (type == null) {
				LOGGER.error("Unknown packet type ${msg::class}!")
			} else if (!type.direction.acceptedOn(side)) {
				LOGGER.error("Packet ${type.type} can not be accepted on side $side!")
			} else {
				try {
					ctx.fireChannelRead(msg)
				} catch (err: Throwable) {
					LOGGER.error("Error while reading incoming packet from network", err)
				}
			}
		}

		override fun write(ctx: ChannelHandlerContext, msg: Any, promise: ChannelPromise) {
			val type = clazz2Type[msg::class]

			if (type == null) {
				LOGGER.error("Unknown outgoing message type ${msg::class}, it will not reach the other side.")
			} else if (!type.direction.acceptedOn(side.opposite)) {
				LOGGER.error("Packet ${type.type} can not be accepted on side ${side.opposite}, refusing to send it!")
			} else {
				ctx.write(msg, promise)
			}
		}
	}

	companion object {
		private val LOGGER = LogManager.getLogger()

		val NATIVE = PacketRegistry(false)
		val LEGACY = PacketRegistry(true)
		val HANDSHAKE = PacketRegistry(false)

		init {
			NATIVE.add(::JoinWorldPacket)
			NATIVE.add(::ChunkCellsPacket)
			NATIVE.add(::ForgetChunkPacket)
			NATIVE.add(::TrackedPositionPacket)
			NATIVE.add(::TrackedSizePacket)
			NATIVE.add(::SpawnWorldObjectPacket)
			NATIVE.add(::ForgetEntityPacket)

			HANDSHAKE.add(::ProtocolRequestPacket)
			HANDSHAKE.add(::ProtocolResponsePacket)

			// legacy protocol handshake looks like this
			// --> ProtocolRequest
			// <-- ProtocolResponse
			// --> ClientConnect
			// <-- HandshakeChallenge *
			// --> HandshakeResponse *
			// <-- ConnectSuccess / ConnectFailure

			LEGACY.skip("ProtocolRequest")
			LEGACY.skip("ProtocolResponse")

			// Packets sent universe server -> universe client
			LEGACY.add(::ServerDisconnectPacket) // ServerDisconnect
			LEGACY.add(::ConnectSuccessPacket) // ConnectSuccess
			LEGACY.skip("ConnectFailure")
			LEGACY.add(::HandshakeChallengePacket) // HandshakeChallenge
			LEGACY.skip("ChatReceive")
			LEGACY.skip("UniverseTimeUpdate")
			LEGACY.skip("CelestialResponse")
			LEGACY.skip("PlayerWarpResult")
			LEGACY.skip("PlanetTypeUpdate")
			LEGACY.skip("Pause")
			LEGACY.skip("ServerInfo")

			// Packets sent universe client -> universe server
			LEGACY.add(::ClientConnectPacket) // ClientConnect
			LEGACY.skip("ClientDisconnectRequest")
			LEGACY.add(::HandshakeResponsePacket) // HandshakeResponse
			LEGACY.skip("PlayerWarp")
			LEGACY.skip("FlyShip")
			LEGACY.skip("ChatSend")
			LEGACY.skip("CelestialRequest")

			// Packets sent bidirectionally between the universe client and the universe
			// server
			LEGACY.skip("ClientContextUpdate")

			// Packets sent world server -> world client
			LEGACY.add(::WorldStartPacket) // WorldStart
			LEGACY.skip("WorldStop")
			LEGACY.skip("WorldLayoutUpdate")
			LEGACY.skip("WorldParametersUpdate")
			LEGACY.skip("CentralStructureUpdate")
			LEGACY.skip("TileArrayUpdate")
			LEGACY.skip("TileUpdate")
			LEGACY.skip("TileLiquidUpdate")
			LEGACY.skip("TileDamageUpdate")
			LEGACY.skip("TileModificationFailure")
			LEGACY.skip("GiveItem")
			LEGACY.skip("EnvironmentUpdate")
			LEGACY.skip("UpdateTileProtection")
			LEGACY.skip("SetDungeonGravity")
			LEGACY.skip("SetDungeonBreathable")
			LEGACY.skip("SetPlayerStart")
			LEGACY.skip("FindUniqueEntityResponse")
			LEGACY.skip("Pong")

			// Packets sent world client -> world server
			LEGACY.skip("ModifyTileList")
			LEGACY.skip("DamageTileGroup")
			LEGACY.skip("CollectLiquid")
			LEGACY.skip("RequestDrop")
			LEGACY.skip("SpawnEntity")
			LEGACY.skip("ConnectWire")
			LEGACY.skip("DisconnectAllWires")
			LEGACY.skip("WorldClientStateUpdate")
			LEGACY.skip("FindUniqueEntity")
			LEGACY.skip("WorldStartAcknowledge")
			LEGACY.skip("Ping")

			// Packets sent bidirectionally between world client and world server
			LEGACY.skip("EntityCreate")
			LEGACY.skip("EntityUpdateSet")
			LEGACY.skip("EntityDestroy")
			LEGACY.skip("EntityInteract")
			LEGACY.skip("EntityInteractResult")
			LEGACY.skip("HitRequest")
			LEGACY.skip("DamageRequest")
			LEGACY.skip("DamageNotification")
			LEGACY.skip("EntityMessage")
			LEGACY.skip("EntityMessageResponse")
			LEGACY.skip("UpdateWorldProperties")
			LEGACY.skip("StepUpdate")

			// Packets sent system server -> system client
			LEGACY.skip("SystemWorldStart")
			LEGACY.skip("SystemWorldUpdate")
			LEGACY.skip("SystemObjectCreate")
			LEGACY.skip("SystemObjectDestroy")
			LEGACY.skip("SystemShipCreate")
			LEGACY.skip("SystemShipDestroy")

			// Packets sent system client -> system server
			LEGACY.skip("SystemObjectSpawn")
		}
	}
}
